from base64 import b64encode
import calendar
from datetime import datetime
import hmac
from hashlib import sha512


def nonce():
    utctime = datetime.utcnow()
    return int(calendar.timegm(utctime.timetuple()) * 10 ** 3 + utctime.microsecond * 10 ** -3)


class HMAC(object):
    def __init__(self, secret, url, method, body=u'', nonce=None):
        self.secret = secret
        self.url = url
        self.method = method
        self.body = body if body else u''
        self._nonce = nonce
        self._signature = None

    @property
    def nonce(self):
        if not self._nonce:
            self._nonce = nonce()
        return self._nonce

    @property
    def signature(self):
        if not self._signature:
            message = u"{}\n{}\n{}\n{}".format(self.nonce, self.method, self.url, self.body)
            digest = hmac.new(self.secret, msg=message, digestmod=sha512)
            self._signature = b64encode(digest.digest()).replace(u'\n', u'')
        return self._signature



from requests.auth import AuthBase

from .HMAC import HMAC


class AmbisafeAuth(AuthBase):
    def __init__(self, api_key, api_secret):
        self.api_key = api_key
        self.api_secret = api_secret

    def __call__(self, request):
        """
        :param request:
        :type request: requests.PreparedRequest
        :return:
        """
        hmac = HMAC(self.api_secret, request.url, request.method, request.body)
        request.headers['Timestamp'] = hmac.nonce
        request.headers['Signature'] = hmac.signature
        request.headers['API-key'] = self.api_key
        return request



import logging
import random
from uuid import uuid4
from pycoin import ecdsa
from pycoin.key import Key
from pycoin.tx.script import der
from ambisafe.crypt import Crypt

logger = logging.getLogger('ambisafe')
SIGHASH_ALL = 1
order = ecdsa.generator_secp256k1.order()


class Container(object):
    def __init__(self, public_key, data, iv, salt):
        """
        Creates container.
        :param public_key: Public key
        :param data: Encrypted private_key
        :param iv: IV
        :param salt: Salt
        :return: Container
        """
        self.publicKey = public_key
        self.data = data
        self.iv = iv
        self.salt = salt

    @classmethod
    def generate_key_pair(cls):
        key = Key(random.SystemRandom().randrange(1, order))
        private_key = format(key.secret_exponent(), 'x')
        if len(private_key) % 2:
            private_key = '0' + private_key
        public_key = key.sec_as_hex()
        return private_key, public_key

    @classmethod
    def generate(cls, secret):
        """
        Generating new container encrypted with secret
        :param secret:
        :type secret: basestring
        :return: Container
        """
        private_key, public_key = cls.generate_key_pair()
        crypt = Crypt(secret)
        salt = str(uuid4())
        try:
            iv, encrypted_private_key = crypt.encrypt(private_key, salt)
        except TypeError:
            print private_key
            raise
        return cls(public_key, encrypted_private_key, iv, salt)

    @classmethod
    def from_server_response(cls, publicKey, data, iv, salt):
        """
        Creating Container object from Ambisafe KeyServer response
        :param publicKey:
        :param data:
        :param iv:
        :param salt:
        :return:
        """
        return cls(publicKey, data, iv, salt)

    def get_private_key(self, secret):
        """
        Get decrypted private key from Container using secret
        :param secret:
        :return: str
        """
        crypt = Crypt(secret)
        private_key = crypt.decrypt(self.data, self.salt, self.iv)
        if not private_key:
            raise ValueError('Wrong secret')
        return private_key

    def sign(self, message, private_key):
        """
        Sign message with private key
        :param message:
        :param private_key:
        :return: str
        """
        key = Key(int(private_key, 16))
        r, s = ecdsa.sign(ecdsa.generator_secp256k1, key.secret_exponent(), int(message, 16))
        if s + s > order:
            s = order - s
        sig = der.sigencode_der(r, s) + chr(SIGHASH_ALL)
        return sig.encode('hex')

    def __getitem__(self, item):
        return self.__dict__[item]

    def as_response(self):
        """
        Get container dict in response format
        :return: dict
        """
        return self.__dict__

    def as_request(self):
        """
        Get container dict in request format
        :return: dict
        """
        container = self.__dict__
        # work around different request and response formats
        container['public_key'] = container.pop('publicKey')
        return container

    def __repr__(self):
        return u'<Container public_key="{}" data="{}" iv="{}" salt="{}">'.format(
            self.publicKey, self.data, self.iv, self.salt
        )



from functools import partial
import logging
import json
from .containers import Container
from .transactions import GrantTransaction

logger = logging.getLogger('ambisafe')


class Account(object):
    @classmethod
    def from_server_response(cls, response):
        """
        Creates Account subclass for corresponding security schema from response from Ambisafe KeyServer
        :param response:
        :type response: dict
        :returns: Account -- Account subclass
        """
        logger.debug('Ambisafe.account.Account.from_server_response | creating account from response: {}'
                     .format(response))
        account_id = response['account']['externalId']
        security_schema = response['account']['securitySchemaName']

        address = response['account']['address']
        if security_schema in ('Wallet4', 'Wallet4v2'):
            containers = {}
            for role, container in response.get('containers', {}).items():
                try:
                    del container['role']
                except KeyError:
                    pass
                containers[role] = Container.from_server_response(**container)
            if security_schema == 'Wallet4':
                version = 1
            else:
                version = 2
            return Wallet4Account(account_id, user_container=containers['USER'],
                                  operator_container=containers['OPERATOR'],
                                  address=address, version=version)
        elif security_schema == 'Simple':
            return SimpleAccount(account_id, address)
        elif security_schema == 'CurrencyIssuer':
            return CurrencyIssuerAccount(account_id, response['containers'], address=address)
        else:
            raise NotImplementedError('This security schema is not implemented')

    def as_dict(self):
        """
        Returns dict representations of account
        :return: dict
        """
        raise NotImplemented

    def to_json(self):
        """
        Returns json representations of account
        :return: str
        """
        return json.dumps(self.as_dict())

    def __repr__(self):
        return '<Account id={} schema={} address={}>'.format(self.id, self._security_schema, self.address)


class SimpleAccount(Account):
    def __init__(self, account_id, address=None, currency='BTC'):
        """
        Base account class
        :param account_id: ID of account
        :type account_id: int or basestring
        :param address:
        :type address: basestring
        :param currency:
        :type currency: basestring
        :return:
        """
        self.id = account_id
        self.currency = currency
        self.address = address
        self._security_schema = 'Simple'

    def as_dict(self):
        """
        Returns dict representations of account
        :return: dict
        """
        return {
            'id': self.id,
            'currency': self.currency,
            'security_schema': self._security_schema,
        }


class Wallet4Account(Account):
    def __init__(self, account_id, operator_container=None,
                 user_container=None, address=None, currency='BTC',
                 version=1):
        """
        Creates Wallet4 account object
        :param account_id:
        :type account_id: basestring or int
        :param operator_container:
        :type operator_container: Container
        :param user_container:
        :type user_container: Container
        :param address:
        :type address: basestring
        :param currency:
        :type currency: basestring
        :return:
        """
        assert isinstance(user_container, Container)
        assert isinstance(operator_container, Container)
        self.user_container = user_container
        self.operator_container = operator_container
        self.id = account_id
        self.currency = currency
        self.address = address
        if version == 1:
            self._security_schema = 'Wallet4'
        elif version == 2:
            self._security_schema = 'Wallet4v2'
        else:
            raise ValueError('Wrong Wallet4 version')

    def as_dict(self):
        """
        Returns dict representations of account
        :return: dict
        """
        return {
            'id': self.id,
            'currency': self.currency,
            'security_schema': self._security_schema,
            'containers': {
                'USER': self.user_container,
                'OPERATOR': self.operator_container,
            },
        }

    def sign(self, transaction, container_name, secret):
        if container_name == 'USER':
            container = self.user_container
        elif container_name == 'OPERATOR':
            container = self.operator_container
        else:
            raise ValueError('container_name should be "USER" or "OPERATOR"')

        private_key = container.get_private_key(secret)
        sign = partial(container.sign, private_key=private_key)
        signed_sighashes = map(sign, transaction.sighashes)

        if container_name == 'USER':
            transaction.user_signatures = signed_sighashes
        elif container_name == 'OPERATOR':
            transaction.operator_signatures = signed_sighashes

        return transaction


class CurrencyIssuerAccount(Account):
    def __init__(self, account_id, containers, address, currency=None):
        self._security_schema = 'CurrencyIssuer'
        self.account_id = account_id
        self.address = address
        self.containers = containers
        self.currency = currency

    def sign(self, transaction, container_number, secret):
        if not isinstance(transaction, GrantTransaction):
            raise ValueError('transaction should be instance of GrantTransaction class')
        container = self.containers[str(container_number)]
        private_key = container.get_private_key(secret)
        sign = partial(container.sign, private_key=private_key)
        signed_sighashes = map(sign, transaction.sighashes)
        transaction.add_signature(container_number, signed_sighashes)



import logging
import json
import re
from pycoin.encoding import is_valid_wif
import requests

from .account import Account
from .auth import AmbisafeAuth
from .containers import Container
from .exc import ClientError, ServerError
from .transactions import BaseTransaction, Wallet4Transaction, RecoveryTransaction, GrantTransaction

logger = logging.getLogger('ambisafe')


class Client(object):
    def __init__(self, ambisafe_server, secret, api_key, api_secret, account_id_prefix='',
                 connect_timeout=None, read_timeout=None):
        if not re.match(r'^(http|https)://', ambisafe_server):
            raise ValueError('ambisafe_server should be http/https URI')
        self.ambisafe_server = ambisafe_server
        self.secret = secret
        self.api_key = api_key
        self.api_secret = api_secret
        self.account_id_prefix = account_id_prefix + ':' if account_id_prefix else ''
        self.connect_timeout = connect_timeout
        self.read_timeout = read_timeout

    def get_prefixed_account_id(self, account_id):
        return '{}{}'.format(self.account_id_prefix, account_id)

    def remove_prefix_from_account_id(self, account_id):
        if self.account_id_prefix:
            return account_id[len(self.account_id_prefix):]
        return account_id

    def create_simple_account(self, account_id, currency='BTC', wif=None):
        """
        Create account with Simple security schema
        :param account_id: int|str
        :param currency: str
        :return:
        """
        logger.debug('Creating account with Simple security schema: account_id: {}, currency: {}'
                     .format(account_id, currency))

        data = {'id': self.get_prefixed_account_id(account_id), 'currency': currency, 'security_schema': 'Simple'}

        if wif is not None:
            if not is_valid_wif(wif):
                raise ValueError('Wrong WIF')
            data['containers'] = {'user': {'wifPk': wif}}

        response = self.make_request('POST', '/accounts', json.dumps(data))

        return Account.from_server_response(response)

    def create_wallet4_account(self, account_id, user_container, operator_container, currency='BTC', version=1):
        """
        Create account with Wallet4 security schema
        :param account_id: int|str
        :param user_container: ambisafe.account.Container
        :param operator_container: ambisafe.account.Container
        :param currency: str
        :return:
        """
        if version == 1:
            security_schema = 'Wallet4'
        elif version == 2:
            security_schema = 'Wallet4v2'
        else:
            raise ValueError('Wrong Wallet4 version')

        logger.debug(
            'Creating account with {} security schema: account_id: {}, currency: {} '
            'containers: (user: {}, operator: {})'
            .format(
                security_schema,
                self.get_prefixed_account_id(account_id),
                currency,
                user_container,
                operator_container
            )
        )

        containers = {
            "USER": user_container.as_request(),
            "OPERATOR": operator_container.as_request(),
        }
        response = self.make_request('POST', '/accounts', json.dumps({
            'id': self.get_prefixed_account_id(account_id),
            'currency': currency,
            'security_schema': security_schema,
            'containers': containers,
        }))

        return Account.from_server_response(response)

    def create_currency_issuer_account(self, account_id, signatures_required, containers, currency):
        logger.debug('Creating account with CurrencyIssuer security schema: account_id: {}, currency: {} '
                     'containers: {} signatures_required: {}'
                     .format(self.get_prefixed_account_id(account_id), currency, containers, signatures_required))
        if any(not isinstance(container, Container) for container in containers):
            raise ValueError("containers should be a list of Container instances")

        containers = [dict(role=str(i), **container.as_request()) for i, container in enumerate(containers)]
        response = self.make_request('POST', '/accounts', json.dumps({
            'id': self.get_prefixed_account_id(account_id),
            'currency': currency,
            'security_schema': 'CurrencyIssuer',
            'signatures_required': signatures_required,
            'containers': containers,
        }))
        return Account.from_server_response(response)

    def update_wallet4_account(self, account_id, user_container, operator_container,
                               currency='BTC', regenerate_server_container=False, version=1):
        if version == 1:
            security_schema = 'Wallet4'
        elif version == 2:
            security_schema = 'Wallet4v2'
        else:
            raise ValueError('Wrong Wallet4 version')
        containers = {
            "USER": user_container.as_request(),
            "OPERATOR": operator_container.as_request(),
        }
        return Account.from_server_response(self.make_request('PUT', '/accounts', json.dumps({
            'id': self.get_prefixed_account_id(account_id),
            'currency': currency,
            'security_schema': security_schema,
            'containers': containers,
            'regenerate_server_container': regenerate_server_container,
        })))

    def get_balance(self, account_id, currency='BTC'):
        balance = self.make_request(
            'GET',
            '/balances/{currency}/{external_id}'
            .format(currency=currency, external_id=self.get_prefixed_account_id(account_id))
        )['balance']
        if balance.startswith('0x'):
            if balance == '0x':
                return 0
            return int(balance, 16)
        return float(balance)

    def get_balance_by_address(self, address, currency):
        balance = self.make_request(
            'GET',
            '/balances/{currency}/address/{address}'
            .format(currency=currency, address=address)
        )['balance']
        if balance.startswith('0x'):
            if balance == '0x':
                return 0
            return int(balance, 16)
        return float(balance)

    def get_account(self, account_id, currency='BTC'):
        account_data = self.make_request('GET', '/accounts/{external_id}/{currency}'
                                         .format(external_id=self.get_prefixed_account_id(account_id),
                                                 currency=currency))
        account_data['account']['externalId'] = self.remove_prefix_from_account_id(
            account_data['account']['externalId']
        )
        return Account.from_server_response(account_data)

    def get_account_by_address(self, address):
        account_data = self.make_request('GET', '/accounts/{address}'
                                         .format(address=address))
        account_data['account']['externalId'] = self.remove_prefix_from_account_id(
            account_data['account']['externalId']
        )
        return Account.from_server_response(account_data)

    def build_transaction(self, account_id, currency, destination, amount):
        body = {
            "destination": destination,
            "amount": amount,
        }
        return Wallet4Transaction(**self.make_request('POST',
                                                      '/transactions/build/{external_id}/{currency}'
                                                      .format(external_id=self.get_prefixed_account_id(account_id),
                                                              currency=currency),
                                                      body=json.dumps(body)))

    def submit(self, account_id, transaction, currency):
        if not isinstance(transaction, BaseTransaction):
            raise ValueError('transaction should be instance of BaseTransaction class')
        return self.make_request('POST', '/transactions/submit/{external_id}/{currency}'
                                 .format(external_id=self.get_prefixed_account_id(account_id), currency=currency),
                                 body=transaction.to_json())

    def sign_wallet4_transaction(self, transaction, account_id, currency):
        account = self.get_account(account_id, currency)
        return account.sign(transaction, 'OPERATOR', self.secret)

    def cosign_wallet4_and_submit(self, transaction, account_id, currency='BTC'):
        transaction = self.sign_wallet4_transaction(transaction, account_id, currency)
        return self.submit(account_id, transaction, currency)

    def build_recovery_transaction(self, account_id, currency, old_address):
        response = self.make_request(
            'POST',
            '/transactions/build_recovery/{external_id}/{currency}/{address}'
            .format(external_id=self.get_prefixed_account_id(account_id), currency=currency, address=old_address)
        )
        logger.debug('response: {}'.format(response))
        return RecoveryTransaction(response['operator'], response['recovery_transaction'], response['account_id'])

    def cosign_and_recovery(self, transaction, account_id, currency='BTC'):
        transaction = self.sign_wallet4_transaction(transaction, account_id, currency)
        return self.submit(account_id, transaction, currency)

    def build_grant_transaction(self, account_id, currency, destination, amount, signatures_required):
        body = {
            "destination": destination,
            "amount": str(amount),
            "signatures_required": str(signatures_required),
            "operation_type": "grant"
        }
        transaction = self.make_request('POST',
                                        '/transactions/build/{id}/{currency}'
                                        .format(id=self.get_prefixed_account_id(account_id), currency=currency),
                                        json.dumps(body))
        return GrantTransaction(**transaction)

    def make_request(self, method, uri, body=None):
        url = self.ambisafe_server + uri
        headers = {u'Accept': u'application/json'}
        if method in [u'POST', u'PUT']:
            headers[u'Content-Type'] = u'application/json'
        logger.debug(u'Request to ambisafe KeyServer: method: {}, url: "{}", headers: {}, data: {}'
                     .format(method, url, headers, body))
        response = requests.request(method, url, headers=headers, data=body,
                                    auth=AmbisafeAuth(self.api_key, self.api_secret),
                                    timeout=(self.connect_timeout, self.read_timeout))
        logger.debug(u'Response from ambisafe KeyServer: status: {}, text: {}'
                     .format(response.status_code, response.text))
        return Client._handle_response(response)

    @staticmethod
    def _handle_response(response):
        try:
            response_data = response.json()
        except ValueError as e:
            # ValueError is parent of JSONDecodeError
            raise ServerError(e.message, '')

        if not response.ok:
            if 400 <= response.status_code < 500:
                raise ClientError(response_data['message'], response_data['error'])
            elif 500 <= response.status_code < 600:
                raise ServerError(response_data['message'], response_data['error'])

        return response_data



from hashlib import pbkdf2_hmac

from Crypto import Random
from Crypto.Cipher import AES
import binascii

ITERATIONS = 1000
KEY_LENGTH = 32
BLOCK_SIZE = AES.block_size

pad = lambda s: s + (BLOCK_SIZE - len(s) % BLOCK_SIZE) * chr(BLOCK_SIZE - len(s) % BLOCK_SIZE)
unpad = lambda s: s[:-ord(s[len(s)-1:])]


def generate_iv():
    return Random.get_random_bytes(BLOCK_SIZE)


class Crypt(object):
    def __init__(self, password):
        self.password = password

    def _derive_key(self, salt):
        return pbkdf2_hmac('sha512', self.password, salt, ITERATIONS, KEY_LENGTH)

    def encrypt(self, data, salt):
        aes = AES.new(self._derive_key(salt), AES.MODE_CBC,
                      generate_iv())
        binary = binascii.unhexlify(data)
        return binascii.hexlify(aes.IV), binascii.hexlify(aes.encrypt(pad(binary)))

    def decrypt(self, encrypted, salt, iv):
        encrypted, salt, iv = str(encrypted), str(salt), str(iv)
        aes = AES.new(self._derive_key(salt), AES.MODE_CBC, iv.decode('hex'))
        return binascii.hexlify(unpad(aes.decrypt(binascii.unhexlify(encrypted))))



import json
import logging
from ambisafe.containers import Container

logger = logging.getLogger('ambisafe')


class BaseTransaction(object):
    def __init__(self, hex, fee, sighashes):
        self.hex = hex
        self.fee = fee
        self.sighashes = sighashes

    def to_json(self):
        raise NotImplemented


class Wallet4Transaction(BaseTransaction):
    def __init__(self, hex, fee, sighashes, operator_signatures=None, user_signatures=None):
        super(Wallet4Transaction, self).__init__(hex, fee, sighashes)
        self.operator_signatures = operator_signatures or []
        self.user_signatures = user_signatures or []

    def to_dict(self):
        return {
            'hex': self.hex,
            'fee': self.fee,
            'sighashes': self.sighashes,
            'operator_signatures': self.operator_signatures,
            'user_signatures': self.user_signatures,
        }

    def to_json(self):
        return json.dumps(self.to_dict())


class RecoveryTransaction(Wallet4Transaction):
    def __init__(self, operator, recovery_transaction, account_id, recovery_address=None):
        self.account_id = account_id
        self.recovery_address = recovery_address
        self.operator_container = Container(operator['publicKey'], operator['data'],
                                            operator['iv'], operator['salt'])
        super(RecoveryTransaction, self).__init__(**recovery_transaction)

    def to_dict(self):
        return {
            'operator': self.operator_container.__dict__,
            'recovery_transaction': {
                'hex': self.hex,
                'fee': self.fee,
                'sighashes': self.sighashes,
            },
            'account_id': self.account_id,
            'operator_signatures': self.operator_signatures,
            'operator_backup_signatures': self.user_signatures,
            'recovery_address': self.recovery_address,
        }

    def to_json(self):
        return json.dumps(self.to_dict())


class GrantTransaction(BaseTransaction):
    def __init__(self, hex, fee, sighashes, signatures=None):
        super(GrantTransaction, self).__init__(hex, fee, sighashes)
        self.signatures = signatures or []

    def add_signature(self, signature_number, sigs):
        if any(signature for signature in self.signatures if signature['key'] == str(signature_number)):
            raise ValueError('Sigs for {} signature_number already exists'.format(signature_number))

        self.signatures.append({
            'key': signature_number,
            'sigs': sigs,
        })

    def to_dict(self):
        return {
            'hex': self.hex,
            'fee': self.fee,
            'sighashes': self.sighashes,
            'signatures': self.signatures,
        }

    def to_json(self):
        return json.dumps(self.to_dict())
